%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Interface to the CLP(R) solver library of Sicstus-Prolog
%
% The clauses in this file are added to each compiled Curry program
% if real arithmetic constraints are used.

:- use_module(library(clpr)).

?- block prim_clpr_plus(?,?,?,-,?).
prim_clpr_plus(RA,RB,R,E0,E) :- hnf(RA,A,E0,E1), hnf(RB,B,E1,E2),
	freeze(E2,({R = A+B}, E2=E)).

?- block prim_clpr_minus(?,?,?,-,?).
prim_clpr_minus(RA,RB,R,E0,E) :- hnf(RA,A,E0,E1), hnf(RB,B,E1,E2),
	freeze(E2,({R = A-B}, E2=E)).

?- block prim_clpr_times(?,?,?,-,?).
prim_clpr_times(RA,RB,R,E0,E) :- hnf(RA,A,E0,E1), hnf(RB,B,E1,E2),
	freeze(E2,({R = A*B}, E2=E)).

?- block prim_clpr_div(?,?,?,-,?).
prim_clpr_div(RA,RB,R,E0,E) :- hnf(RA,A,E0,E1), hnf(RB,B,E1,E2),
	freeze(E2,({R = A/B}, E2=E)).

?- block prim_clpr_le(?,?,?,-,?).
prim_clpr_le(RA,RB,success,E0,E) :- hnf(RA,A,E0,E1), hnf(RB,B,E1,E2),
	freeze(E2,({A<B}, E2=E)).

?- block prim_clpr_ge(?,?,?,-,?).
prim_clpr_ge(RA,RB,success,E0,E) :- hnf(RA,A,E0,E1), hnf(RB,B,E1,E2),
	freeze(E2,({A>B}, E2=E)).

?- block prim_clpr_leq(?,?,?,-,?).
prim_clpr_leq(RA,RB,success,E0,E) :- hnf(RA,A,E0,E1), hnf(RB,B,E1,E2),
	freeze(E2,({A=<B}, E2=E)).

?- block prim_clpr_geq(?,?,?,-,?).
prim_clpr_geq(RA,RB,success,E0,E) :- hnf(RA,A,E0,E1), hnf(RB,B,E1,E2),
	freeze(E2,({A>=B}, E2=E)).

% transform an integer into a float:
?- block prim_clpr_i2f(?,?,-,?).
prim_clpr_i2f(RA,R,E0,E) :-
	constructorRootedAndDeref(RA,A,E0,E1),
	freeze(E1,(R is A*1.0, E1=E)).

?- block prim_minimumFor(?,?,?,-,?).
prim_minimumFor(Guard,Fun,X,E0,E) :-
        waitUntilGround('(,)'(Guard,Fun),E0,E1),
	exec_minimum(Guard,Fun,X,E1,E).

?- block exec_minimum(?,?,?,-,?).
exec_minimum(Guard,Fun,X,E,E3) :-
        prim_apply(Guard,X,success,E,E1),
        prim_apply(Fun,X,Z,E1,E2),
	minimize(Z), E2=E3.

?- block 'prim_CLPR.maximumFor'(?,?,?,-,?).
'prim_CLPR.maximumFor'(Guard,Fun,X,E0,E) :-
        waitUntilGround('(,)'(Guard,Fun),E0,E1),
	exec_maximum(Guard,Fun,X,E1,E).

?- block exec_maximum(?,?,?,-,?).
exec_maximum(Guard,Fun,X,E,E3) :-
        prim_apply(Guard,X,success,E,E1),
        prim_apply(Fun,X,Z,E1,E2),
	maximize(Z), E2=E3.
