%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Definition of external functions of module IO:
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- use_module(library(system)).

?- block prim_stdin(?,-,?).
prim_stdin(Stream,E0,E) :- prolog_flag(user_input,Stream), E0=E.

?- block prim_stdout(?,-,?).
prim_stdout(Stream,E0,E) :- prolog_flag(user_output,Stream), E0=E.

?- block prim_stderr(?,-,?).
prim_stderr(Stream,E0,E) :- prolog_flag(user_error,Stream), E0=E.

?- block prim_openFile(?,?,?,?,-,?).
prim_openFile(RA,RMode,_,'$io'(Stream),E0,E) :-
	derefAll(RA,A), deref(RMode,Mode),
	string2Atom(A,FName),
	curryFileMode2plmode(Mode,PMode),
	open(FName,PMode,Stream), E0=E.

curryFileMode2plmode('IO.ReadMode',read).
curryFileMode2plmode('IO.WriteMode',write).
curryFileMode2plmode('IO.AppendMode',append).

?- block prim_hClose(?,?,?,?,-,?).
prim_hClose(Stream,_,'$io'('()'),E0,E) :-
	hnf(Stream,HStream,E0,E1),
	freeze(E1,((isOutputStream(HStream) -> flush_output(HStream) ; true),
		   close(HStream), E1=E)).

?- block prim_hFlush(?,?,?,?,-,?).
prim_hFlush(Stream,_,'$io'('()'),E0,E) :-
        hnf(Stream,HStream,E0,E1),
	freeze(E1,((isOutputStream(HStream) -> flush_output(HStream) ; true),
		   E1=E)).

?- block prim_hIsEOF(?,?,?,-,?).
prim_hIsEOF(Stream,_,'$io'(B),E0,E) :-
        hnf(Stream,HStream,E0,E1),
	freeze(E1,((at_end_of_stream(HStream) -> B='True' ; B='False'), E1=E)).

?- block prim_hWaitForInput(?,?,?,?,-,?).
prim_hWaitForInput(Stream,RTO,_,R,E0,E) :-
	hnf(Stream,HStream,E0,E1),
	hnf(RTO,TO,E1,E2),
	prim_hWaitForInput_exec(HStream,TO,R,E2,E).

?- block prim_hWaitForInput_exec(-,?,?,?,?),
         prim_hWaitForInput_exec(?,-,?,?,?),
         prim_hWaitForInput_exec(?,?,?,-,?).
prim_hWaitForInput_exec(Stream,TimeOut,'$io'(B),E0,E) :-
	TOSec is TimeOut//1000,
	TOMSec is (TimeOut mod 1000) * 1000, % in milliseconds
	socket_select([],_,TOSec:TOMSec,[Stream],SelStreams),
	(SelStreams = [Stream] -> B='True' ; B='False'),
	!, E0=E.

?- block prim_hGetChar(?,?,?,-,?).
prim_hGetChar(Stream,_,'$io'(C),E0,E) :-
        hnf(Stream,HStream,E0,E1),
	freeze(E1,(get0(HStream,N), char_int(C,N), E1=E)).

?- block prim_hPutChar(?,?,?,?,-,?).
prim_hPutChar(Stream,RC,_,R,E0,E) :-
	hnf(Stream,HStream,E0,E1),
	hnf(RC,C,E1,E2),
	prim_hPutChar_exec(HStream,C,R,E2,E).

?- block prim_hPutChar_exec(-,?,?,?,?), prim_hPutChar_exec(?,-,?,?,?),
         prim_hPutChar_exec(?,?,?,-,?).
prim_hPutChar_exec(Stream,C,'$io'('()'),E0,E) :-
	char_int(C,N), put(Stream,N), E0=E.

?- block prim_hIsReadable(-,?,?,?,?), prim_hIsReadable(?,?,?,-,?).
prim_hIsReadable(Stream,_,'$io'(B),E0,E) :-
        hnf(Stream,HStream,E0,E1),
	freeze(E1,((isInputStream(HStream) -> B='True' ; B='False'), E1=E)).

?- block prim_hIsWritable(-,?,?,?,?), prim_hIsWritable(?,?,?,-,?).
prim_hIsWritable(Stream,_,'$io'(B),E0,E) :-
        hnf(Stream,HStream,E0,E1),
	freeze(E1,((isOutputStream(HStream) -> B='True' ; B='False'), E1=E)).


isInputStream(Stream) :-
	sicstus38orHigher -> stream_property(Stream,input)
	                   ; current_stream(_,input,Stream).
isOutputStream(Stream) :-
	sicstus38orHigher -> stream_property(Stream,output)
	                   ; current_stream(_,output,Stream).
