%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Definitions of builtins of module Term:
%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% conversion of Curry data terms into string representation in
% standard prefix notation

?- block prim_showQTerm(?,?,-,?).
prim_showQTerm(Exp,String,E0,E) :-
	nf(Exp,Term,E0,E1), waitUntilGround(Term,E1,E2),
	prim_showQTerm_exec(Term,String,E2,E).

?- block prim_showQTerm_exec(?,?,-,?).
prim_showQTerm_exec(Term,String,E0,E) :-
	show_term(Term,qualified,String,[]), E0=E.

?- block prim_showTerm(?,?,-,?).
prim_showTerm(Exp,String,E0,E) :-
	nf(Exp,Term,E0,E1), waitUntilGround(Term,E1,E2),
	prim_showTerm_exec(Term,String,E2,E).

?- block prim_showTerm_exec(?,?,-,?).
prim_showTerm_exec(Term,String,E0,E) :-
	show_term(Term,unqualified,String,[]), E0=E.

% with difference lists to speed up:
show_term(V,_,S,S) :-
	var(V), !,
	write(user_error,'*** Internal error in ReadShowTerm.showTerm: free variable'),
	nl(user_error).
% special case for Unsafe.showAny(Q)Term:
show_term('_'(I),_,[Underscore|S],E) :- !,
	char_int(Underscore,95),
	number_chars(I,SI), map2(char_int,String,SI), diffList(String,S,E).
show_term([],_,[OpSqBracket,ClSqBracket|E],E) :- !,
	char_int(OpSqBracket,91), char_int(ClSqBracket,93).
show_term(I,_,S,E) :-
	integer(I), !,
	number_chars(I,SI), map2(char_int,String,SI), diffList(String,S,E).
show_term(F,_,S,E) :-
	float(F), !,
	number_chars(F,SF), map2(char_int,String,SF), diffList(String,S,E).
show_term(C,_,[Apo|S],E) :- % 39='''
	isCharCons(C), !,
	char_int(Apo,39),
	char_int(C,N),
	show_termchar(N,S,SE),
	SE = [Apo|E].
show_term('()',_,[Op,Cl|E],E) :- !, char_int(Op,40), char_int(Cl,41).
show_term('VAR',_,S,E) :-
	!,
	atom2String('VAR',SV), diffList(SV,S,E).
show_term(T,Q,S,E) :-
	atom(T), !,
	atom2String(T,ST),
	(Q=qualified -> ShowT=ST
	              ; char_int(Dot,46), removeQualifier(ST,Dot,ShowT)),
	(isId(T) -> diffList(ShowT,S,E)
	          ; % enclose in parentheses:
	            char_int(Op,40), char_int(Cl,41),
	            append([Op|ShowT],[Cl|E],S)),
	!.
show_term(L,_,[Quot|S],E) :- % 34 = '"'
	isString(L), !,
	char_int(Quot,34),
	show_termstring(L,S,E).
show_term([H|T],Q,[OpSqBracket|SH],E) :-
	isCompleteList(T), !,
	char_int(OpSqBracket,91),
	show_term(H,Q,SH,ST),
	show_termcomplist(T,Q,ST,E).
show_term([H|T],Q,[Op|SH],E) :- !,
	char_int(Op,40), char_int(Cl,41),
	show_termlist([H|T],Q,SH,SHE),
	SHE=[Cl|E].
show_term(T,Q,[Op|SC],E) :- % 40='('
	isShowableArg(T),
	T =.. [C|Args], !,
	char_int(Op,40), char_int(Cl,41),
	(isTupleCons(C) -> show_termtuple(Args,Q,SC,SArgsE)
	 ; show_term(C,Q,SC,SCE),
	   show_termargs(Args,Q,SCE,SArgsE)),
	SArgsE=[Cl|E]. % 41=')'

isTupleCons(S) :- atom_chars(S,[40,44|_]). % tuple constructor

show_termstring([],[Quot|E],E) :- char_int(Quot,34).
show_termstring([C|T],S,E) :-
	char_int(C,N),
	show_termchar(N,S,ST),
	show_termstring(T,ST,E).

show_termchar(34,[C1,C2|E],E) :- !, cp_string([C1,C2],[92,34]).
show_termchar(92,[C1,C2|E],E) :- !, cp_string([C1,C2],[92,92]).
show_termchar(10,[C1,C2|E],E) :- !, cp_string([C1,C2],[92,110]).
show_termchar(13,[C1,C2|E],E) :- !, cp_string([C1,C2],[92,114]).
show_termchar(9,[C1,C2|E],E) :- !, cp_string([C1,C2],[92,116]).
show_termchar(8,[C1,C2|E],E) :- !, cp_string([C1,C2],[92,98]).
show_termchar(N,[C1,C2,C3,C4|E],E) :- (N<32 ; N>126), !,
	N1 is (N//100)+48, N2 is ((N mod 100)//10)+48,
	N3 is (N mod 10)+48,
	cp_string([C1,C2,C3,C4],[92,N1,N2,N3]).
show_termchar(N,[C|E],E) :- char_int(C,N).

show_termlist(L,Q,SH,E) :- nonvar(L), L=[H|T], !,
	show_term(H,Q,SH,SHE),
	char_int(Colon,58), % 58=':'
	SHE=[Colon|ST],
	show_termlist(T,Q,ST,E).
show_termlist(L,Q,SH,E) :-
	show_term(L,Q,SH,E).

show_termcomplist([],_,[ClSqBracket|E],E) :- char_int(ClSqBracket,93).
show_termcomplist([H|T],Q,[Comma|SH],E) :- % 44=','
	char_int(Comma,44),
	show_term(H,Q,SH,ST),
	show_termcomplist(T,Q,ST,E).

show_termargs([],_,S,S).
show_termargs([A|As],Q,[Blank|SA],E) :-
	char_int(Blank,32),
	show_term(A,Q,SA,SAE),
	show_termargs(As,Q,SAE,E).

show_termtuple([A],Q,SA,SAE) :- show_term(A,Q,SA,SAE).
show_termtuple([A1,A2|As],Q,SA,E) :-
	show_term(A1,Q,SA,SA1),
	char_int(Comma,44),
	SA1 = [Comma|SAE], % 44=','
	show_termtuple([A2|As],Q,SAE,E).

% check argument whether it can be encoded as a string:
isShowableArg('Ports.internalPort'(_,SNr,_,_)) :-
	SNr=<0, % is internal port or process port?
	write(user_error,'ERROR: cannot serialize internal port!'),
	nl(user_error),
	!, fail.
isShowableArg(_).

% remove module qualifier from internal name:
removeQualifier(N,Dot,UT) :- removeQualifier(N,Dot,N,UT).
removeQualifier([],_,N,N). % no qualifier in name, keep original name
removeQualifier([C|Cs],Dot,N,UN) :-
	C=Dot -> UN=Cs ; removeQualifier(Cs,Dot,N,UN).

% convert list (arg 1) into difference list (arg 2+3):
diffList([],E,E).
diffList([H|T],[H|DT],DTE) :- diffList(T,DT,DTE).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% conversion of standard prefix string representations of Curry terms
% into Curry terms:

% conversion of string representations of Curry terms into Curry terms:
?- block prim_readsQTerm(?,?,-,?).
prim_readsQTerm(RString,R,E0,E) :-
	nf(RString,String,E0,E1), waitUntilGround(String,E1,E2),
        prim_readsQTerm_exec(String,R,E2,E).

?- block prim_readsQTerm_exec(?,?,-,?).
prim_readsQTerm_exec(String,['(,)'(Term,TailString)],E0,E) :-
	map2(char_int,String,PrologString),
	readTerm(PrologString,qualified,Tail,Term),
	map2(char_int,TailString,Tail), !, E0=E.
prim_readsQTerm_exec(_,[],E,E). % parse error

?- block prim_readsUnqualifiedTerm(?,?,?,-,?).
prim_readsUnqualifiedTerm(RPrefixes,RString,R,E0,E) :-
	nf(RPrefixes,Prefixes,E0,E1), waitUntilGround(Prefixes,E1,E2),
	nf(RString,String,E2,E3), waitUntilGround(String,E3,E4),
        prim_readsUnqualifiedTerm_exec(Prefixes,String,R,E4,E).

?- block prim_readsUnqualifiedTerm_exec(?,?,?,-,?).
prim_readsUnqualifiedTerm_exec(Prefixes,String,['(,)'(Term,TailString)],E0,E) :-
	(Prefixes=[] -> PrefixDots=any
	              ; map2(prefix2prefixdot,Prefixes,PrefixDots)),
	map2(char_int,String,PrologString),
	readTerm(PrologString,unqualified(PrefixDots),Tail,Term),
	map2(char_int,TailString,Tail), !, E0=E.
prim_readsUnqualifiedTerm_exec(_,_,[],E,E). % parse error

prefix2prefixdot(CPrefix,[]) :-
	map2(char_int,CPrefix,"prelude"), !.
prefix2prefixdot(CPrefix,PrefixDot) :-
	map2(char_int,CPrefix,Prefix), append(Prefix,[46],PrefixDot).

readTerm(S,Q,T,Term) :-
	skipWhiteSpace(S,T1),
	readTerm0(T1,Q,T,Term).

readTerm0([C|Cs],Q,T,Term) :-
	isLetter(C), !,
	readTermS([C|Cs],Q,T2,Func),
	skipWhiteSpace(T2,T3), readTermArgs(T3,Q,T,Args),
	Term =.. [Func|Args].
readTerm0(S,Q,T,Term) :-
	readTermS(S,Q,T,Term).

% special case for Unsafe.readAny(Q)Term:
readTermS([95|S],Q,T,'_'(Num)) :- % variable encoding
	(Q=any_qualified ; Q=any_unqualified(_)), !,
	numberconst(NumStr,S,T), !, % use c2p parser
	number_chars(Num,NumStr).
readTermS([C|S],_,T,Num) :- % number
	C>47, C<58,
	numberconst(NumStr,[C|S],T), !, % use c2p parser
	number_chars(Num,NumStr).
readTermS([45,C|S],_,T,Num) :- % negative number
	C>47, C<58,
	numberconst(NumStr,[C|S],T), !, % use c2p parser
	number_chars(PNum,NumStr),
	Num is 0-PNum.
readTermS([91,93|T],_,T,[]) :- !. % empty list
readTermS([91|S],Q,T,L) :- !, readCompList(S,Q,T,L). % non-empty list
readTermS([39|S],_,T,C) :- !, % character
	(readChar(S,T,C) -> true ; readCharParseError([39|S])).
readTermS([34|S],_,T,String) :- !, % string
	readString(S,T,String).
readTermS([40,41|T],_,T,'()') :- !.
readTermS([40|S],Q,T,Term) :- !,	% open bracket
	readTerm(S,Q,SA1,Term1), skipWhiteSpace(SA1,SA2),
	(SA2=[58|SA3] -> readList(SA3,Q,SA4,Tail),
	                 Term=[Term1|Tail], SA4=[41|T]
	 ; (SA2=[44|SA3] -> readTuple(SA3,Q,SA4,Tail), SA4=[41|T],
	                    Args=[Term1|Tail], length(Args,NA),
	                    Cons=")", prefixComma(Cons,NA,ConsComma),
	                    atom_chars(TupleCons,[40|ConsComma]),
	                    Term =.. [TupleCons|Args]
 	                  ; (SA2=[41|T] -> Term=Term1
			     ; readTermArgs(SA2,Q,[41|T],Args),
	                       Term=..[Term1|Args]))).
readTermS(S,Q,T,QId) :-
        (readQVarOpId(S,T,IdString) -> true ; readParseError(S)),
        atom_chars(Id,IdString),
        readIdTerm(Id,Q,S,QId).

readCharParseError(S) :-
	pakcsrc(readtermerrors,yes),
	write(user_error,'ERROR in ReadShowTerm.readTerm: illegal character in remaining string:'),
	nl(user_error),
	putChars(user_error,S), nl(user_error),
	!, fail.

readParseError(S) :-
	pakcsrc(readtermerrors,yes),
	write(user_error,'ERROR in ReadShowTerm.readTerm: cannot parse remaining string:'),
	nl(user_error),
	putChars(user_error,S), nl(user_error),
	!, fail.

readIdTerm('VAR',_,_,'VAR') :- !.
readIdTerm(Id,unchecked,_,Id) :- !. % don't check constructors (for readFlcFromFcy.pl)
readIdTerm(Id,qualified,_,Id) :-
	constructorOrFunctionType(Id,_,_,_), !.
readIdTerm(Id,unqualified(Prefixes),_,QId) :-
	addQualifier(Prefixes,Id,QId).
readIdTerm(Id,any_qualified,_,Id) :-
	constructorOrFunctionType(Id,_,_,_), !.
readIdTerm(Id,any_unqualified(Prefixes),_,QId) :-
	addQualifier(Prefixes,Id,QId).
readIdTerm(Id,_,S,_) :-
	pakcsrc(readtermerrors,yes),
	write(user_error,'ERROR in ReadShowTerm.readTerm: Unknown symbol: '),
	write(user_error,Id), nl(user_error),
	write(user_error,'in remaining term string: '),
	putChars(user_error,S), nl(user_error),
	!, fail.

addQualifier(any,Id,QId) :-
	constructorOrFunctionType(QId,Id,_,_), constructorOrFunctionType(QJ,Id,_,_),
	\+ QId=QJ, !,
	writeErr('WARNING: Unqualified symbol "'), writeErr(Id),
	writeErr('" not unique due to multiple imports.'), nlErr.
addQualifier(any,Id,QId) :-
	constructorOrFunctionType(QId,Id,_,_), !.
addQualifier([Prefix|_],Id,QId) :-
	atom_chars(Id,IdS),
	append(Prefix,IdS,QIdS), atom_chars(QId,QIdS),
	constructorOrFunctionType(QId,_,_,_), !.
addQualifier([_|Prefixes],Id,QId) :- addQualifier(Prefixes,Id,QId).
	
readCompList(S,Q,T,[Elem|Tail]) :-
	readTerm(S,Q,SA1,Elem), skipWhiteSpace(SA1,SA2),
	(SA2=[93|SA3] -> T=SA3, Tail=[] ;
	 SA2=[44|SA3], readCompList(SA3,Q,T,Tail)).

readList(S,Q,T,L) :-
	readTerm(S,Q,SA1,Elem), skipWhiteSpace(SA1,SA2),
	(SA2=[58|SA3] -> L=[Elem|Tail], readList(SA3,Q,T,Tail)
	               ; L=Elem, T=SA2).

readTuple(S,Q,T,L) :-
	readTerm(S,Q,SA1,Elem), skipWhiteSpace(SA1,SA2),
	(SA2=[44|SA3] -> L=[Elem|Tail], readTuple(SA3,Q,T,Tail)
	               ; L=[Elem], T=SA2).

readTermArgs([],_,[],[]) :- !.
readTermArgs([41|S],_,[41|S],[]) :- !. % 41 = )
readTermArgs([44|S],_,[44|S],[]) :- !. % 44 = ,
readTermArgs([93|S],_,[93|S],[]) :- !. % 93 = ]
readTermArgs(S,Q,T,[Arg|Args]) :-
	readTermS(S,Q,SA1,Arg), skipWhiteSpace(SA1,SA2),
	readTermArgs(SA2,Q,T,Args).

readChar([92,N1,N2,N3,39|T],T,C) :- N1>=48, N1<52, !,
	N is (N1-48)*100+(N2-48)*10+N3-48,
	char_int(C,N).
readChar([92,N,39|T],T,C) :- !,
	readStringChar(N,NS),
	char_int(C,NS).
readChar([N,39|T],T,C) :- char_int(C,N).
	
readString([34|T],T,[]) :- !.
readString([92,N1,N2,N3|Ns],T,[C|Str]) :- N1>=48, N1<52, !,
	N is (N1-48)*100+(N2-48)*10+N3-48,
	char_int(C,N),
	readString(Ns,T,Str).
readString([92,N|Ns],T,[C|Str]) :- !,
	readStringChar(N,NS),
	char_int(C,NS),
	readString(Ns,T,Str).
readString([N|Ns],T,[C|Str]) :-
	char_int(C,N),
	readString(Ns,T,Str).
readStringChar(34,34) :- !.
readStringChar(92,92) :- !.
readStringChar(110,10) :- !.
readStringChar(114,13) :- !.
readStringChar(116,9) :- !.
readStringChar(98,8) :- !.
readStringChar(N,N).

readQVarOpId([C|Cs],T,[C|Str]) :-
	isLetter(C) -> readModOrVar(Cs,T,Str) ; readOpId(Cs,T,Str).

readModOrVar([C|Cs],T,[C|Str]) :-
	isVarIdChar(C), !,
	(C=46 -> readQVarOpId(Cs,T,Str) ; readModOrVar(Cs,T,Str)).
readModOrVar(T,T,[]).


readOpId([C|Cs],T,[C|Str]) :-
	isOpIdChar(C), !,
	readOpId(Cs,T,Str).
readOpId(T,T,[]).

skipWhiteSpace([C|Cs],Cs1) :- isWhiteSpace(C), !, skipWhiteSpace(Cs,Cs1).
skipWhiteSpace(Cs,Cs).

isWhiteSpace(32).
isWhiteSpace(10).
isWhiteSpace(13).
isWhiteSpace(12).
isWhiteSpace( 9).

% is a character of a variable identifier, i.e.,
% A..Z a..z 0..9 . _ '
%isVarIdChar(C) :-
%	65=<C, C=<90 ; 97=<C, C=<122 ; 48=<C, C=<57 ; C=46 ; C=95 ; C=39.

% generated by:
% mapIO_ (putStrLn . ("isVarIdChar("++)  . (++").") . show) ([65..90]++[97..122]++[48..57]++[46,95,39])
isVarIdChar(65).
isVarIdChar(66).
isVarIdChar(67).
isVarIdChar(68).
isVarIdChar(69).
isVarIdChar(70).
isVarIdChar(71).
isVarIdChar(72).
isVarIdChar(73).
isVarIdChar(74).
isVarIdChar(75).
isVarIdChar(76).
isVarIdChar(77).
isVarIdChar(78).
isVarIdChar(79).
isVarIdChar(80).
isVarIdChar(81).
isVarIdChar(82).
isVarIdChar(83).
isVarIdChar(84).
isVarIdChar(85).
isVarIdChar(86).
isVarIdChar(87).
isVarIdChar(88).
isVarIdChar(89).
isVarIdChar(90).
isVarIdChar(97).
isVarIdChar(98).
isVarIdChar(99).
isVarIdChar(100).
isVarIdChar(101).
isVarIdChar(102).
isVarIdChar(103).
isVarIdChar(104).
isVarIdChar(105).
isVarIdChar(106).
isVarIdChar(107).
isVarIdChar(108).
isVarIdChar(109).
isVarIdChar(110).
isVarIdChar(111).
isVarIdChar(112).
isVarIdChar(113).
isVarIdChar(114).
isVarIdChar(115).
isVarIdChar(116).
isVarIdChar(117).
isVarIdChar(118).
isVarIdChar(119).
isVarIdChar(120).
isVarIdChar(121).
isVarIdChar(122).
isVarIdChar(48).
isVarIdChar(49).
isVarIdChar(50).
isVarIdChar(51).
isVarIdChar(52).
isVarIdChar(53).
isVarIdChar(54).
isVarIdChar(55).
isVarIdChar(56).
isVarIdChar(57).
isVarIdChar(46).
isVarIdChar(95).
isVarIdChar(39).

isOpIdChar(126).% ~
isOpIdChar(33).	% !
isOpIdChar(64).	% @
isOpIdChar(35).	% #
isOpIdChar(36).	% $
isOpIdChar(37).	% %
isOpIdChar(94).	% ^
isOpIdChar(38).	% &
isOpIdChar(42).	% *
isOpIdChar(43).	% +
isOpIdChar(45).	% -
isOpIdChar(61).	% =
isOpIdChar(60).	% <
isOpIdChar(62).	% >
isOpIdChar(63).	% ?
isOpIdChar(46).	% .
isOpIdChar(47).	% /
isOpIdChar(124).% |
isOpIdChar(92).	% \
isOpIdChar(58).	% :

isLetter(65).
isLetter(66).
isLetter(67).
isLetter(68).
isLetter(69).
isLetter(70).
isLetter(71).
isLetter(72).
isLetter(73).
isLetter(74).
isLetter(75).
isLetter(76).
isLetter(77).
isLetter(78).
isLetter(79).
isLetter(80).
isLetter(81).
isLetter(82).
isLetter(83).
isLetter(84).
isLetter(85).
isLetter(86).
isLetter(87).
isLetter(88).
isLetter(89).
isLetter(90).
isLetter(97).
isLetter(98).
isLetter(99).
isLetter(100).
isLetter(101).
isLetter(102).
isLetter(103).
isLetter(104).
isLetter(105).
isLetter(106).
isLetter(107).
isLetter(108).
isLetter(109).
isLetter(110).
isLetter(111).
isLetter(112).
isLetter(113).
isLetter(114).
isLetter(115).
isLetter(116).
isLetter(117).
isLetter(118).
isLetter(119).
isLetter(120).
isLetter(121).
isLetter(122).

