%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Implementation of the database primitives.
%
% The clauses in this file are added to each compiled Curry program
% if the database primitives (module: DB) are used

:- use_module(library(db)).

:- dynamic connectedDB/3. % already opened databases (name/dbref/status
                          % where status is open or close)

?- block 'prim_TermDB.openDB'(-,?,?,?,?), 'prim_TermDB.openDB'(?,?,?,-,?).
'prim_TermDB.openDB'(Name,_,'$io'('TermDB_DBRef'(DB)),E0,E) :-
	string2Atom(Name,DB),
	(connectedDB(DB,_,open)
	 -> true
	  ; (retract(connectedDB(DB,_,_)) -> true ; true),
	    db_open(DB,update,DBRef),
	    assertz(connectedDB(DB,DBRef,open))),
	E0=E.

?- block 'prim_TermDB.closeDB'(-,?,?,?,?), 'prim_TermDB.closeDB'(?,?,?,-,?).
'prim_TermDB.closeDB'('TermDB_DBRef'(DB),_,'$io'('()'),E0,E) :-
	(retract(connectedDB(DB,DBRef,open))
	  -> db_close(DBRef), assertz(connectedDB(DB,DBRef,close))
	   ; true), E0=E.

getCurrentDBRef(DB,DBRef) :-
	connectedDB(DB,DBRef,open)
	 -> true
	  ; (retract(connectedDB(DB,_,_)) -> true ; true),
	    db_open(DB,update,DBRef),
	    assertz(connectedDB(DB,DBRef,open)).

?- block 'prim_TermDB.storeDB'(?,-,?,?,?,?),
         'prim_TermDB.storeDB'(?,?,?,?,-,?).
'prim_TermDB.storeDB'(Term,'TermDB_DBRef'(DB),_,'$io'('()'),E0,E) :-
        getCurrentDBRef(DB,DBRef),
	db_store(DBRef,Term,_), E0=E.

?- block 'prim_TermDB.eraseDB'(?,-,?,?,?,?),
         'prim_TermDB.eraseDB'(?,?,?,?,-,?).
'prim_TermDB.eraseDB'(Term,'TermDB_DBRef'(DB),_,Result,E0,E) :-
        getCurrentDBRef(DB,DBRef),
	(db_fetch(DBRef,Term,TermRef)
	  -> db_erase(DBRef,TermRef),
	     Result='$io'('True')
	   ; Result='$io'('False')),
	!,
	E0=E.

?- block 'prim_TermDB.findallDB'(?,-,?,?,?,?),
         'prim_TermDB.findallDB'(?,?,?,?,-,?).
'prim_TermDB.findallDB'(Term,'TermDB_DBRef'(DB),_,Result,E0,E) :-
        getCurrentDBRef(DB,DBRef),
	db_findall(DBRef,Term,Terms),
	Result='$io'(Terms),
	E0=E.

