%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Definitions of builtins of module Unsafe:
%

:- use_module(prim_readshowterm).

?- block 'prim_unsafePerformIO'(?,?,-,?).
'prim_unsafePerformIO'(Action,H,E0,E) :-
        worldToken(World),
        prim_apply(Action,World,'$io'(Result),E0,E1),
	hnf(Result,H,E1,E).

?- block 'prim_spawnConstraint'(?,?,?,-,?).
'prim_spawnConstraint'(Guard,Exp,H,E0,E) :-
        hnf(Guard,S,E0,_), S=success,
	hnf(Exp,H,E0,E).

?- block 'prim_isVar'(?,?,-,?).
'prim_isVar'(RTerm,H,E0,E) :- deref(RTerm,Term),
        (var(Term) -> H='True' ; H='False'), E0=E.

?- block 'prim_identicalVar'(?,?,?,-,?).
'prim_identicalVar'(RX,RY,H,E0,E) :- deref(RX,X), deref(RY,Y),
        ((var(X), var(Y), X==Y) -> H='True' ; H='False'), E0=E.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% conversion of Curry data terms (with variables) into string representation in
% standard prefix notation

?- block prim_showAnyTerm(?,?,-,?).
prim_showAnyTerm(Exp,String,E0,E) :-
	nf(Exp,Term,E0,E1),
	prim_showAnyTerm_exec(Term,String,E1,E).

?- block prim_showAnyTerm_exec(?,?,-,?).
prim_showAnyTerm_exec(Term,String,E0,E) :-
        copy_term(Term,CTerm),
	groundTermVars(CTerm,0,_),
	show_term(CTerm,unqualified,String,[]), E0=E.

?- block prim_showAnyQTerm(?,?,-,?).
prim_showAnyQTerm(Exp,String,E0,E) :-
	nf(Exp,Term,E0,E1),
	prim_showAnyQTerm_exec(Term,String,E1,E).

?- block prim_showAnyQTerm_exec(?,?,-,?).
prim_showAnyQTerm_exec(Term,String,E0,E) :-
        copy_term(Term,CTerm),
	groundTermVars(CTerm,0,_),
	show_term(CTerm,qualified,String,[]), E0=E.

?- block prim_showAnyExpression(?,?,-,?).
prim_showAnyExpression(Exp,String,E0,E) :-
        removeShares(Exp,UExp), copy_term(UExp,CExp),
	groundTermVars(CExp,0,_),
	show_term(CExp,unqualified,String,[]), E0=E.

?- block prim_showAnyExpression(?,?,-,?).
prim_showAnyQExpression(Exp,String,E0,E) :-
        removeShares(Exp,UExp), copy_term(UExp,CExp),
	groundTermVars(CExp,0,_),
	show_term(CExp,qualified,String,[]), E0=E.


% bind free variables in a term to a printable ground representation:
groundTermVars(X,I,I1) :- var(X), !,
	X='_'(I),
	I1 is I+1.
groundTermVars(A,I,I) :- atom(A), !.
groundTermVars(T,I,I1) :-
	T =.. [_|Args],
	groundTermsVars(Args,I,I1).

groundTermsVars([],I,I).
groundTermsVars([A|As],I,I2) :-
	groundTermVars(A,I,I1),
	groundTermsVars(As,I1,I2).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% conversion of standard prefix string representations of Curry terms
% into Curry terms:

% conversion of string representations of Curry terms into Curry terms:
?- block prim_readsAnyQTerm(?,?,-,?).
prim_readsAnyQTerm(RString,R,E0,E) :-
	nf(RString,String,E0,E1), waitUntilGround(String,E1,E2),
        prim_readsAnyQTerm_exec(String,R,E2,E).

?- block prim_readsAnyQTerm_exec(?,?,-,?).
prim_readsAnyQTerm_exec(String,['(,)'(Term,TailString)],E0,E) :-
	map2(char_int,String,PrologString),
	readTerm(PrologString,any_qualified,Tail,GTerm),
	ungroundTermVars(GTerm,Term,_),
	map2(char_int,TailString,Tail), !, E0=E.
prim_readsAnyQTerm_exec(_,[],E,E). % parse error

?- block prim_readsAnyUnqualifiedTerm(?,?,?,-,?).
prim_readsAnyUnqualifiedTerm(RPrefixes,RString,R,E0,E) :-
	nf(RPrefixes,Prefixes,E0,E1), waitUntilGround(Prefixes,E1,E2),
	nf(RString,String,E2,E3), waitUntilGround(String,E3,E4),
        prim_readsAnyUnqualifiedTerm_exec(Prefixes,String,R,E4,E).

?- block prim_readsAnyUnqualifiedTerm_exec(?,?,?,-,?).
prim_readsAnyUnqualifiedTerm_exec(Prefixes,String,['(,)'(Term,TailString)],E0,E) :-
	(Prefixes=[] -> PrefixDots=any
	              ; map2(prefix2prefixdot,Prefixes,PrefixDots)),
	map2(char_int,String,PrologString),
	readTerm(PrologString,any_unqualified(PrefixDots),Tail,GTerm),
	ungroundTermVars(GTerm,Term,_),
	map2(char_int,TailString,Tail), !, E0=E.
prim_readsAnyUnqualifiedTerm_exec(_,_,[],E,E). % parse error


% replace ground representations by free variables in a term:
ungroundTermVars('_'(I),X,Binds) :- !,
	getVarIndex(Binds,I,X).
ungroundTermVars(A,A,_) :- atom(A), !.
ungroundTermVars(T,VT,Binds) :-
	T =.. [C|Args],
	ungroundTermsVars(Args,VArgs,Binds),
	VT =.. [C|VArgs].

ungroundTermsVars([],[],_).
ungroundTermsVars([A|As],[VA|VAs],Binds) :-
	ungroundTermVars(A,VA,Binds),
	ungroundTermsVars(As,VAs,Binds).

getVarIndex(Binds,Idx,Var) :- var(Binds), !, Binds=[(Idx=Var)|_].
getVarIndex([(Idx=V)|_],Idx,Var) :- !, V=Var.
getVarIndex([_|Binds],Idx,Var) :- getVarIndex(Binds,Idx,Var).

