-- Simple graph visualization
module ShowGraph(viewDependencyGraph) where

import IO
import IOExts
import Char(isAlphaNum)
import List(intersperse)

---------------------------------------------------------------------------------
-- Required software (adjust to your local installation):

-- The "dot" graph drawing tool:
dotCmd = "dot"

-- The program to view for Postscript files passed to stdin:
psviewCmd = "kghostview -"

---------------------------------------------------------------------------------

-- Show a dependency graph with dot.
-- A dependency graph consists of a list of triples of the form (n,as,ms),
-- where n is a node name, as (dot) attributes for node n, and ms the list
-- of direct dependents from n.
viewDependencyGraph :: [(String,[(String,String)],[String])] -> IO ()
viewDependencyGraph deps = viewDot (deps2dot deps)

-- translate dependencies into DOT language:
deps2dot :: [(String,[(String,String)],[String])] -> String
deps2dot deps =
  "digraph dependencies{\n" ++ concatMap dep2dot deps ++ "}\n"
 where
  dep2dot (x,attrs,xdeps) =
    let attrtxt = if null attrs then ""
                  else showDotID x ++
                       '[':concat (intersperse ","
                                    (map (\ (n,v)->n++"=\""++v++"\"") attrs)) ++ "]"
                        ++ ";\n"  in
    if null xdeps
    then if null attrs then showDotID x ++ ";\n" else attrtxt
    else concatMap (\i->showDotID x ++ " -> " ++ showDotID i ++ ";\n") xdeps ++ attrtxt

showDotID s | all isAlphaNum s = s
            | otherwise = '"' : concatMap escapeDQ s ++ "\""
 where escapeDQ c = if c=='"' then "\\\"" else [c]

-- visualize a DOT string:
viewDot dottxt = 
    --writeFile "/home/mh/tmp.dot" dottxt >>
    execCmd (dotCmd++" -Tps | "++psviewCmd) >>= \ (dotIn,dotOut,dotErr) ->
    hPutStr dotIn dottxt >> hFlush dotIn >>
    hClose dotIn >> hClose dotOut >> hClose dotErr 
