-- Linearity optimizer for non-strict equality:

import FlatCurry
import FlatCurryRead
import Linearity
import Dependency
import FileGoodies
import System
import List(intersperse)

m = optimizeNonstrictEqualityInModule "last"

-- Check arguments and call main function:
main = do
  args <- getArgs
  case args of
    [prog] -> optimizeNonstrictEqualityInModule (stripSuffix prog)
    _ -> putStrLn $ "ERROR: Illegal arguments: " ++
                    concat (intersperse " " args) ++ "\n" ++
                    "Usage: OptNonStrict.state <module_name>"

optimizeNonstrictEqualityInModule prog = do
  progs <- readFlatCurryWithImports prog
  writeFCY (prog++"_lo.fcy")
           (optimizeNonstrictEquality (concatMap funcsOfProg progs) (head progs))
  system ("mv "++prog++"_lo.fcy "++prog++".fcy")
  putStrLn ("Optimized program written to '"++prog++".fcy'")

-- Replace "=:<=" by "=:<<=" if the left argument is a linear term with
-- linear functions.
-- The first argument is the list of all functions occuring in the program,
-- the second argument is the current module to be optimized.
optimizeNonstrictEquality :: [FuncDecl] -> Prog -> Prog
optimizeNonstrictEquality allfuns (Prog mod imps ts funs ops) =
 Prog mod imps ts (map (optimizeFun (analyseRightLinearity allfuns)) funs) ops

optimizeFun lininfo (Func qn ar vis ty (Rule vs e)) =
   Func qn ar vis ty (Rule vs (optimizeExp lininfo e))
optimizeFun _ (Func qn ar vis ty (External e)) = Func qn ar vis ty (External e)

optimizeExp _ (Var i) = Var i
optimizeExp _ (Lit l) = Lit l
optimizeExp lininfo (Comb ct f es)
 | f==("prelude","=:<=") && length es == 2 && linearExpr (head es)
   && onlyLinearFunctions lininfo (head es)
  = Comb ct ("prelude","=:<<=") es
 | otherwise = Comb ct f (map (optimizeExp lininfo) es)
optimizeExp lininfo (Free vs e) = Free vs (optimizeExp lininfo e)
optimizeExp li (Let bs exp) =
   Let (map (\(v,e)->(v,optimizeExp li e)) bs) (optimizeExp li exp)
optimizeExp lininfo (Or e1 e2) = Or (optimizeExp lininfo e1) (optimizeExp lininfo e2)
optimizeExp li (Case ct e bs) = Case ct (optimizeExp li e) (map optimizeBranch bs)
 where
  optimizeBranch (Branch patt be) = Branch patt (optimizeExp li be)

onlyLinearFunctions :: [(QName,Bool)] -> Expr -> Bool
onlyLinearFunctions li e = all isRightLinearDefined (funcsInExpr e)
 where
   isRightLinearDefined fun = maybe False id (lookup fun li)

-- goodies:

funcsOfProg (Prog _ _ _ funcs _) = funcs
